"""Unit tests for to_networkx/graph_builder.py."""

from uuid import uuid4

import pytest

from infoextract_cidoc.io.to_networkx.graph_builder import (
    build_graph_from_entities,
    create_temporal_graph,
    to_networkx_graph,
)
from infoextract_cidoc.models.base import CRMEntity, CRMRelation
from infoextract_cidoc.models.generated.e_classes import (
    E5_Event,
    E22_HumanMadeObject,
    E52_TimeSpan,
    E53_Place,
)


@pytest.mark.unit
class TestToNetworkxGraph:
    def test_creates_nodes_for_entities(self) -> None:
        entities = [
            CRMEntity(id=uuid4(), class_code="E21", label="Einstein"),
            CRMEntity(id=uuid4(), class_code="E53", label="Princeton"),
        ]
        graph = to_networkx_graph(entities)
        assert graph.number_of_nodes() == 2

    def test_creates_edges_for_relationships(self) -> None:
        e1 = CRMEntity(id=uuid4(), class_code="E21", label="Einstein")
        e2 = CRMEntity(id=uuid4(), class_code="E53", label="Princeton")
        rel = CRMRelation(src=e1.id, type="P74", tgt=e2.id)
        graph = to_networkx_graph([e1, e2], [rel])
        assert graph.number_of_edges() == 1

    def test_node_attributes_stored(self) -> None:
        entity = CRMEntity(id=uuid4(), class_code="E21", label="Einstein")
        graph = to_networkx_graph([entity])
        node_data = graph.nodes[str(entity.id)]
        assert node_data["class_code"] == "E21"
        assert node_data["label"] == "Einstein"

    def test_empty_entities(self) -> None:
        graph = to_networkx_graph([])
        assert graph.number_of_nodes() == 0


@pytest.mark.unit
class TestBuildGraphFromEntities:
    def test_expands_shortcut_fields(self) -> None:
        place = E53_Place(id=uuid4(), class_code="E53", label="Athens")
        timespan = E52_TimeSpan(id=uuid4(), class_code="E52", label="450 BCE")
        event = E5_Event(
            id=uuid4(),
            class_code="E5",
            label="Vase Creation",
            timespan=timespan.id,
            took_place_at=place.id,
        )
        graph = build_graph_from_entities([event, place, timespan])
        # Should have 2 edges from P4 and P7 shortcut expansion
        assert graph.number_of_edges() == 2

    def test_no_shortcuts_when_disabled(self) -> None:
        place = E53_Place(id=uuid4(), class_code="E53", label="Athens")
        event = E5_Event(
            id=uuid4(), class_code="E5", label="Event", took_place_at=place.id
        )
        graph = build_graph_from_entities([event, place], expand_shortcuts=False)
        assert graph.number_of_edges() == 0

    def test_no_self_loops_by_default(self) -> None:
        import networkx as nx

        # Create entity where current_location points to its own ID
        own_id = uuid4()
        entity = E22_HumanMadeObject(
            id=own_id, class_code="E22", label="Vase", current_location=own_id
        )
        graph = build_graph_from_entities([entity], include_self_loops=False)
        assert nx.number_of_selfloops(graph) == 0


@pytest.mark.unit
class TestCreateTemporalGraph:
    def test_marks_temporal_nodes(self) -> None:
        timespan = E52_TimeSpan(id=uuid4(), class_code="E52", label="1879")
        event = E5_Event(
            id=uuid4(),
            class_code="E5",
            label="Birth",
            timespan=timespan.id,
        )
        temporal_graph = create_temporal_graph([event, timespan], [])
        # Check that nodes have has_temporal_info attribute set
        node_data = dict(temporal_graph.nodes(data=True))
        # Event has a timespan shortcut field — so its node_data will have "timespan" key
        event_node = node_data.get(str(event.id), {})
        assert "has_temporal_info" in event_node

    def test_empty_graph(self) -> None:
        temporal_graph = create_temporal_graph([], [])
        assert temporal_graph.number_of_nodes() == 0



"""
Domain/range typing validation for CIDOC CRM.
Validates that relationships align with property domain/range constraints.
"""

import logging
from typing import Any

from infoextract_cidoc.models.base import (
    CRMEntity,
    CRMValidationError,
    CRMValidationWarning,
)
from infoextract_cidoc.models.shortcuts import P_TO_FIELD
from infoextract_cidoc.properties import DOMAIN, P

from .quantifiers import ValidationSeverity

logger = logging.getLogger(__name__)


def validate_domain_range_alignment(
    source_entity: CRMEntity,
    target_entity: CRMEntity,
    p_code: str,
    severity: ValidationSeverity = ValidationSeverity.WARN,
) -> None:
    """
    Validate that a relationship aligns with domain/range constraints.

    Args:
        source_entity: Source entity of the relationship
        target_entity: Target entity of the relationship
        p_code: P-property code for the relationship
        severity: How to handle violations

    Raises:
        CRMValidationError: If severity is RAISE and alignment is violated
        CRMValidationWarning: If severity is WARN and alignment is violated
    """
    if severity == ValidationSeverity.IGNORE:
        return

    if p_code not in P:
        logger.warning("Unknown property code: %s", p_code)
        return

    property_info = P[p_code]
    expected_domain = str(property_info["domain"])
    expected_range = str(property_info["range"])

    # Check domain alignment
    if not _is_class_compatible(source_entity.class_code, expected_domain):
        message = f"Entity {source_entity.id} (class {source_entity.class_code}) does not match domain {expected_domain} for property {p_code}"
        _handle_violation(message, severity, source_entity, target_entity, p_code)

    # Check range alignment
    if not _is_class_compatible(target_entity.class_code, expected_range):
        message = f"Entity {target_entity.id} (class {target_entity.class_code}) does not match range {expected_range} for property {p_code}"
        _handle_violation(message, severity, source_entity, target_entity, p_code)


def validate_entity_typing(
    entity: CRMEntity,
    entity_lookup: dict[str, CRMEntity],
    severity: ValidationSeverity = ValidationSeverity.WARN,
) -> list[str]:
    """
    Validate typing rules for an entity against a lookup table.

    Args:
        entity: The CRM entity to validate
        entity_lookup: Dictionary mapping entity IDs to entities
        severity: How to handle violations

    Returns:
        List of validation messages
    """
    messages = []

    # Get all properties that apply to this entity's class
    applicable_properties = DOMAIN.get(entity.class_code, [])

    for p_code in applicable_properties:
        try:
            # Get target entity IDs for this property
            target_ids = _get_property_target_ids(entity, p_code)

            for target_id in target_ids:
                if target_id in entity_lookup:
                    target_entity = entity_lookup[target_id]
                    validate_domain_range_alignment(
                        entity, target_entity, p_code, severity
                    )
                else:
                    logger.info(
                        "Target entity %s not found in lookup for property %s",
                        target_id,
                        p_code,
                    )

        except (CRMValidationError, CRMValidationWarning) as e:
            messages.append(str(e))
        except Exception as e:
            logger.exception("Error validating typing for property %s", p_code)
            messages.append(f"Error validating typing for property {p_code}: {e}")

    return messages


def validate_batch_typing(
    entities: list[CRMEntity], severity: ValidationSeverity = ValidationSeverity.WARN
) -> dict[str, list[str]]:
    """
    Validate typing rules for a batch of entities.

    Args:
        entities: List of CRM entities to validate
        severity: How to handle violations

    Returns:
        Dictionary mapping entity IDs to validation messages
    """
    # Create entity lookup
    entity_lookup = {str(entity.id): entity for entity in entities}

    results = {}

    for entity in entities:
        messages = validate_entity_typing(entity, entity_lookup, severity)
        if messages:
            results[str(entity.id)] = messages

    return results


def _is_class_compatible(entity_class: str, expected_class: str) -> bool:
    """
    Check if an entity class is compatible with an expected class.

    Args:
        entity_class: The entity's class code
        expected_class: The expected class code

    Returns:
        True if compatible, False otherwise
    """
    # Direct match
    if entity_class == expected_class:
        return True

    # Check inheritance hierarchy
    # This is a simplified version - in practice, you'd want to load
    # the full inheritance hierarchy from the YAML specs
    inheritance_map = {
        "E2": ["E1"],
        "E3": ["E2", "E1"],
        "E4": ["E2", "E1"],
        "E5": ["E2", "E1"],
        "E6": ["E5", "E2", "E1"],
        "E7": ["E5", "E2", "E1"],
        "E8": ["E7", "E5", "E2", "E1"],
        "E9": ["E7", "E5", "E2", "E1"],
        "E10": ["E7", "E5", "E2", "E1"],
        "E11": ["E7", "E5", "E2", "E1"],
        "E12": ["E7", "E5", "E2", "E1"],
        "E13": ["E7", "E5", "E2", "E1"],
        "E14": ["E13", "E7", "E5", "E2", "E1"],
        "E15": ["E13", "E7", "E5", "E2", "E1"],
        "E16": ["E13", "E7", "E5", "E2", "E1"],
        "E17": ["E13", "E7", "E5", "E2", "E1"],
        "E18": ["E1"],
        "E19": ["E18", "E1"],
        "E20": ["E19", "E18", "E1"],
        "E21": ["E20", "E19", "E18", "E1"],
        "E22": ["E19", "E18", "E1"],
        "E23": ["E1"],
        "E24": ["E18", "E1"],
        "E25": ["E24", "E18", "E1"],
        "E26": ["E18", "E1"],
        "E27": ["E26", "E18", "E1"],
        "E28": ["E23", "E1"],
        "E29": ["E28", "E23", "E1"],
        "E30": ["E28", "E23", "E1"],
        "E31": ["E28", "E23", "E1"],
        "E32": ["E31", "E28", "E23", "E1"],
        "E33": ["E28", "E23", "E1"],
        "E34": ["E33", "E28", "E23", "E1"],
        "E35": ["E33", "E28", "E23", "E1"],
        "E36": ["E28", "E23", "E1"],
        "E37": ["E36", "E28", "E23", "E1"],
        "E38": ["E36", "E28", "E23", "E1"],
        "E39": ["E1"],
        "E40": ["E39", "E1"],
        "E41": ["E28", "E23", "E1"],
        "E42": ["E28", "E23", "E1"],
        "E43": ["E1"],
        "E44": ["E41", "E28", "E23", "E1"],
        "E45": ["E44", "E41", "E28", "E23", "E1"],
        "E46": ["E43", "E1"],
        "E47": ["E28", "E23", "E1"],
        "E48": ["E44", "E41", "E28", "E23", "E1"],
        "E49": ["E41", "E28", "E23", "E1"],
        "E50": ["E49", "E41", "E28", "E23", "E1"],
        "E51": ["E45", "E44", "E41", "E28", "E23", "E1"],
        "E52": ["E1"],
        "E53": ["E43", "E1"],
        "E54": ["E28", "E23", "E1"],
        "E55": ["E28", "E23", "E1"],
        "E56": ["E55", "E28", "E23", "E1"],
        "E57": ["E55", "E28", "E23", "E1"],
        "E58": ["E55", "E28", "E23", "E1"],
        "E59": ["E1"],
        "E60": ["E59", "E1"],
        "E61": ["E59", "E1"],
        "E62": ["E59", "E1"],
        "E63": ["E5", "E2", "E1"],
        "E64": ["E5", "E2", "E1"],
        "E65": ["E63", "E5", "E2", "E1"],
        "E66": ["E63", "E5", "E2", "E1"],
        "E67": ["E66", "E63", "E5", "E2", "E1"],
        "E68": ["E64", "E5", "E2", "E1"],
        "E69": ["E64", "E5", "E2", "E1"],
        "E70": ["E1"],
        "E71": ["E70", "E1"],
        "E72": ["E70", "E1"],
        "E73": ["E70", "E1"],
        "E74": ["E39", "E1"],
        "E75": ["E41", "E28", "E23", "E1"],
        "E76": ["E42", "E28", "E23", "E1"],
        "E77": ["E1"],
        "E78": ["E77", "E1"],
        "E79": ["E11", "E7", "E5", "E2", "E1"],
        "E80": ["E11", "E7", "E5", "E2", "E1"],
        "E81": ["E11", "E7", "E5", "E2", "E1"],
        "E82": ["E75", "E41", "E28", "E23", "E1"],
        "E83": ["E65", "E63", "E5", "E2", "E1"],
        "E84": ["E73", "E70", "E1"],
        "E85": ["E7", "E5", "E2", "E1"],
        "E86": ["E7", "E5", "E2", "E1"],
        "E87": ["E7", "E5", "E2", "E1"],
        "E88": ["E28", "E23", "E1"],
        "E89": ["E88", "E28", "E23", "E1"],
        "E90": ["E28", "E23", "E1"],
        "E91": ["E90", "E28", "E23", "E1"],
        "E92": ["E1"],
        "E93": ["E92", "E1"],
        "E94": ["E92", "E1"],
        "E95": ["E59", "E1"],
        "E96": ["E8", "E7", "E5", "E2", "E1"],
        "E97": ["E28", "E23", "E1"],
        "E98": ["E55", "E28", "E23", "E1"],
        "E99": ["E55", "E28", "E23", "E1"],
    }

    # Check if expected_class is in the inheritance chain
    inheritance_chain = inheritance_map.get(entity_class, [])
    return expected_class in inheritance_chain


def _get_property_target_ids(entity: CRMEntity, p_code: str) -> list[str]:
    """
    Get target entity IDs for a property from an entity.

    Args:
        entity: The CRM entity
        p_code: P-property code

    Returns:
        List of target entity IDs
    """
    field_name = P_TO_FIELD.get(p_code)
    if not field_name:
        return []

    if hasattr(entity, field_name):
        value = getattr(entity, field_name)
        if value is None:
            return []
        if isinstance(value, list):
            return [str(v) for v in value]
        return [str(value)]

    return []


def _handle_violation(
    message: str,
    severity: ValidationSeverity,
    source_entity: CRMEntity,
    target_entity: CRMEntity,
    p_code: str,
) -> None:
    """
    Handle a typing violation based on severity.

    Args:
        message: Violation message
        severity: How to handle the violation
        source_entity: Source entity of the relationship
        target_entity: Target entity of the relationship
        p_code: The property code that was violated
    """
    full_message = f"{message} (Property: {p_code}, Source: {source_entity.id}, Target: {target_entity.id})"

    if severity == ValidationSeverity.RAISE:
        raise CRMValidationError(full_message)
    if severity == ValidationSeverity.WARN:
        logger.warning(full_message)
        # Also issue a warning that can be caught
        import warnings

        warnings.warn(full_message, CRMValidationWarning, stacklevel=2)
    # IGNORE is handled in the calling function


def get_typing_summary(entities: list[CRMEntity]) -> dict[str, Any]:
    """
    Get a summary of typing validation results.

    Args:
        entities: List of CRM entities to analyze

    Returns:
        Summary dictionary with validation statistics
    """
    total_entities = len(entities)
    validation_results = validate_batch_typing(entities, ValidationSeverity.WARN)

    entities_with_issues = len(validation_results)
    total_issues = sum(len(messages) for messages in validation_results.values())

    return {
        "total_entities": total_entities,
        "entities_with_issues": entities_with_issues,
        "total_issues": total_issues,
        "validation_results": validation_results,
    }



"""Unit tests for entity and relationship resolution pipeline."""

import pytest

from infoextract_cidoc.extraction.lite_schema import (
    LiteEntity,
    LiteExtractionResult,
    LiteRelationship,
)
from infoextract_cidoc.extraction.resolution import EntityRegistry, resolve_extraction


@pytest.mark.unit
class TestEntityRegistry:
    def test_register_entity(self) -> None:
        registry = EntityRegistry()
        lite = LiteEntity(ref_id="person_1", entity_type="Person", label="Einstein")
        entity = registry.register(lite)
        assert entity.label == "Einstein"
        assert entity.class_code == "E21"

    def test_stable_uuid(self) -> None:
        registry1 = EntityRegistry()
        registry2 = EntityRegistry()
        lite = LiteEntity(ref_id="person_1", entity_type="Person", label="Einstein")
        e1 = registry1.register(lite)
        e2 = registry2.register(lite)
        assert e1.id == e2.id  # deterministic UUID5

    def test_deduplication_by_label(self) -> None:
        registry = EntityRegistry()
        lite1 = LiteEntity(ref_id="person_1", entity_type="Person", label="Einstein")
        lite2 = LiteEntity(ref_id="person_2", entity_type="Person", label="Einstein")
        e1 = registry.register(lite1)
        e2 = registry.register(lite2)
        assert e1.id == e2.id  # same entity
        assert len(registry.entities) == 1

    def test_resolve_by_ref_id(self) -> None:
        registry = EntityRegistry()
        lite = LiteEntity(ref_id="place_1", entity_type="Place", label="Ulm")
        registry.register(lite)
        resolved = registry.resolve("place_1")
        assert resolved is not None
        assert resolved.label == "Ulm"

    def test_resolve_missing_returns_none(self) -> None:
        registry = EntityRegistry()
        assert registry.resolve("nonexistent") is None


@pytest.mark.unit
class TestResolveExtraction:
    def test_basic_resolution(self) -> None:
        lite_result = LiteExtractionResult(
            entities=[
                LiteEntity(ref_id="person_1", entity_type="Person", label="Einstein"),
                LiteEntity(ref_id="place_1", entity_type="Place", label="Ulm"),
            ],
            relationships=[
                LiteRelationship(
                    source_ref="person_1",
                    target_ref="place_1",
                    property_code="P98",
                    property_label="was born",
                ),
            ],
        )
        result = resolve_extraction(lite_result)
        assert len(result.entities) == 2
        assert len(result.relationships) == 1

    def test_broken_source_ref_excluded(self) -> None:
        lite_result = LiteExtractionResult(
            entities=[
                LiteEntity(ref_id="place_1", entity_type="Place", label="Ulm"),
            ],
            relationships=[
                LiteRelationship(
                    source_ref="nonexistent",
                    target_ref="place_1",
                    property_code="P98",
                    property_label="was born",
                ),
            ],
        )
        result = resolve_extraction(lite_result)
        assert len(result.entities) == 1
        assert len(result.relationships) == 0  # broken link excluded

    def test_broken_target_ref_excluded(self) -> None:
        lite_result = LiteExtractionResult(
            entities=[
                LiteEntity(ref_id="person_1", entity_type="Person", label="Einstein"),
            ],
            relationships=[
                LiteRelationship(
                    source_ref="person_1",
                    target_ref="nonexistent",
                    property_code="P98",
                    property_label="was born",
                ),
            ],
        )
        result = resolve_extraction(lite_result)
        assert len(result.relationships) == 0

    def test_entity_type_mapping(self) -> None:
        lite_result = LiteExtractionResult(
            entities=[
                LiteEntity(ref_id="person_1", entity_type="Person", label="P"),
                LiteEntity(ref_id="event_1", entity_type="Event", label="E"),
                LiteEntity(ref_id="place_1", entity_type="Place", label="Pl"),
                LiteEntity(ref_id="object_1", entity_type="Object", label="O"),
                LiteEntity(ref_id="time_1", entity_type="TimeSpan", label="T"),
            ],
        )
        result = resolve_extraction(lite_result)
        codes = {e.class_code for e in result.entities}
        assert codes == {"E21", "E5", "E53", "E22", "E52"}



"""Shared test fixtures for infoextract-cidoc tests."""

from uuid import uuid4

import pytest

from infoextract_cidoc.models.base import CRMEntity, CRMRelation
from infoextract_cidoc.models.generated.e_classes import (
    E5_Event,
    E12_Production,
    E21_Person,
    E22_HumanMadeObject,
    E52_TimeSpan,
    E53_Place,
)


@pytest.fixture
def person_entity() -> E21_Person:
    return E21_Person(id=uuid4(), class_code="E21", label="Albert Einstein")


@pytest.fixture
def place_entity() -> E53_Place:
    return E53_Place(id=uuid4(), class_code="E53", label="Princeton")


@pytest.fixture
def event_entity(place_entity: E53_Place, timespan_entity: "E52_TimeSpan") -> E5_Event:
    return E5_Event(
        id=uuid4(),
        class_code="E5",
        label="Nobel Prize",
        timespan=timespan_entity.id,
        took_place_at=place_entity.id,
    )


@pytest.fixture
def timespan_entity() -> E52_TimeSpan:
    return E52_TimeSpan(id=uuid4(), class_code="E52", label="1921")


@pytest.fixture
def object_entity(
    place_entity: E53_Place, production_entity: E12_Production
) -> E22_HumanMadeObject:
    return E22_HumanMadeObject(
        id=uuid4(),
        class_code="E22",
        label="Ancient Vase",
        current_location=place_entity.id,
        produced_by=production_entity.id,
    )


@pytest.fixture
def production_entity(
    timespan_entity: E52_TimeSpan, place_entity: E53_Place
) -> E12_Production:
    return E12_Production(
        id=uuid4(),
        class_code="E12",
        label="Vase Production",
        timespan=timespan_entity.id,
        took_place_at=place_entity.id,
    )


@pytest.fixture
def sample_entities(
    person_entity: E21_Person,
    place_entity: E53_Place,
    event_entity: E5_Event,
) -> list[CRMEntity]:
    return [person_entity, place_entity, event_entity]


@pytest.fixture
def sample_relations(
    person_entity: E21_Person,
    place_entity: E53_Place,
) -> list[CRMRelation]:
    return [
        CRMRelation(src=person_entity.id, type="P74", tgt=place_entity.id),
    ]



"""
Unit tests for Markdown renderers.
"""

from uuid import uuid4

import pytest

from ...io.to_markdown import MarkdownStyle, render_table, to_markdown
from ...models.generated.e_classes import (
    E12_Production,
    E22_HumanMadeObject,
)


@pytest.mark.unit
class TestMarkdownRendering:
    """Test Markdown rendering functionality."""

    def test_card_rendering(self):
        """Test card-style rendering."""
        entity = E22_HumanMadeObject(
            id=uuid4(),
            class_code="E22",
            label="Ancient Vase",
            type=["E55:Vessel"],
            current_location=uuid4(),
            produced_by=uuid4(),
        )

        markdown = to_markdown(entity, MarkdownStyle.CARD)

        # Check that key elements are present
        assert "### E22 · Human-Made Object · Ancient Vase" in markdown
        assert "(" in markdown and ")" in markdown  # Should have UUID in parentheses
        assert "**Location** (`current_location`):" in markdown
        assert "**Produced By** (`produced_by`):" in markdown

    def test_detailed_rendering(self):
        """Test detailed-style rendering."""
        entity = E22_HumanMadeObject(
            id=uuid4(),
            class_code="E22",
            label="Ancient Vase",
            type=["E55:Vessel"],
            notes="A beautiful amphora",
        )

        markdown = to_markdown(entity, MarkdownStyle.DETAILED)

        # Check that key elements are present
        assert "## E22 · Human-Made Object — Ancient Vase" in markdown
        assert "(" in markdown and ")" in markdown  # Should have UUID in parentheses
        assert "**Label** (`label`): Ancient Vase" in markdown
        assert "**Notes** (`notes`): A beautiful amphora" in markdown

    def test_table_rendering(self):
        """Test table-style rendering."""
        entities = [
            E22_HumanMadeObject(id=uuid4(), class_code="E22", label="Vase 1"),
            E22_HumanMadeObject(id=uuid4(), class_code="E22", label="Vase 2"),
        ]

        markdown = render_table(entities)

        # Check that table structure is present
        assert "| id | class_code | label | type |" in markdown
        assert "| --- | --- | --- | --- |" in markdown
        assert "| E22 | Vase 1 |" in markdown
        assert "| E22 | Vase 2 |" in markdown

    def test_narrative_rendering(self):
        """Test narrative-style rendering."""
        entity = E12_Production(
            id=uuid4(),
            class_code="E12",
            label="Vase Production",
            timespan=uuid4(),
            took_place_at=uuid4(),
        )

        markdown = to_markdown(entity, MarkdownStyle.NARRATIVE)

        # Check that narrative elements are present
        assert "**Vase Production**" in markdown
        assert "is a production" in markdown
        assert "that occurred during" in markdown
        assert "at" in markdown

    def test_aliases_usage(self):
        """Test that aliases are used when provided."""
        entity = E22_HumanMadeObject(id=uuid4(), class_code="E22", label="Ancient Vase")

        aliases = {"E22": "Artifact", "current_location": "Location"}

        markdown = to_markdown(entity, MarkdownStyle.CARD, aliases=aliases)

        # Check that aliases are used
        assert "### E22 · Artifact · Ancient Vase" in markdown
        assert "(" in markdown and ")" in markdown  # Should have UUID in parentheses

    def test_show_codes_option(self):
        """Test show_codes option."""
        entity = E22_HumanMadeObject(id=uuid4(), class_code="E22", label="Ancient Vase")

        # With codes
        markdown_with_codes = to_markdown(entity, MarkdownStyle.CARD, show_codes=True)
        assert "**Label** (`label`): Ancient Vase" in markdown_with_codes

        # Without codes
        markdown_without_codes = to_markdown(
            entity, MarkdownStyle.CARD, show_codes=False
        )
        assert "**Label**: Ancient Vase" in markdown_without_codes
        assert "(`label`)" not in markdown_without_codes

    def test_empty_entity(self):
        """Test rendering of entity with minimal data."""
        entity = E22_HumanMadeObject(id=uuid4(), class_code="E22")

        markdown = to_markdown(entity, MarkdownStyle.CARD)

        # Should still render basic structure
        assert "### E22 · Human-Made Object" in markdown
        assert "(" in markdown and ")" in markdown  # Should have UUID in parentheses

    def test_custom_columns(self):
        """Test table rendering with custom columns."""
        entities = [
            E22_HumanMadeObject(id=uuid4(), class_code="E22", label="Vase 1"),
            E22_HumanMadeObject(id=uuid4(), class_code="E22", label="Vase 2"),
        ]

        custom_columns = ["id", "label"]
        markdown = render_table(entities, columns=custom_columns)

        # Check that only custom columns are present
        assert "| id | label |" in markdown
        assert "class_code" not in markdown
        assert "type" not in markdown

    def test_empty_entities_list(self):
        """Test table rendering with empty entities list."""
        markdown = render_table([])
        assert "No entities to display." in markdown



"""Unit tests for LiteExtractionResult schema."""

import pytest
from pydantic import ValidationError

from infoextract_cidoc.extraction.lite_schema import (
    LiteEntity,
    LiteExtractionResult,
    LiteRelationship,
)


@pytest.mark.unit
class TestLiteEntity:
    def test_valid_entity(self) -> None:
        entity = LiteEntity(
            ref_id="person_1",
            entity_type="Person",
            label="Albert Einstein",
            confidence=0.95,
        )
        assert entity.ref_id == "person_1"
        assert entity.label == "Albert Einstein"
        assert entity.confidence == 0.95

    def test_confidence_bounds(self) -> None:
        with pytest.raises(ValidationError):
            LiteEntity(
                ref_id="person_1",
                entity_type="Person",
                label="Test",
                confidence=1.5,  # out of range
            )

    def test_defaults(self) -> None:
        entity = LiteEntity(
            ref_id="place_1",
            entity_type="Place",
            label="Warsaw",
        )
        assert entity.confidence == 1.0
        assert entity.description is None
        assert entity.attributes == {}


@pytest.mark.unit
class TestLiteRelationship:
    def test_valid_relationship(self) -> None:
        rel = LiteRelationship(
            source_ref="person_1",
            target_ref="place_1",
            property_code="P98",
            property_label="was born",
            confidence=0.9,
        )
        assert rel.source_ref == "person_1"
        assert rel.property_code == "P98"


@pytest.mark.unit
class TestLiteExtractionResult:
    def test_empty_result(self) -> None:
        result = LiteExtractionResult()
        assert result.entities == []
        assert result.relationships == []
        assert result.overall_confidence == 1.0

    def test_with_entities_and_relationships(self) -> None:
        result = LiteExtractionResult(
            entities=[
                LiteEntity(ref_id="person_1", entity_type="Person", label="Einstein"),
                LiteEntity(ref_id="place_1", entity_type="Place", label="Ulm"),
            ],
            relationships=[
                LiteRelationship(
                    source_ref="person_1",
                    target_ref="place_1",
                    property_code="P98",
                    property_label="was born",
                ),
            ],
        )
        assert len(result.entities) == 2
        assert len(result.relationships) == 1



"""CRM validation framework for infoextract-cidoc."""

from .quantifiers import (
    ValidationSeverity,
    enforce_quantifier,
    validate_batch_quantifiers,
    validate_entity_quantifiers,
)
from .typing_rules import validate_batch_typing, validate_domain_range_alignment

__all__ = [
    "ValidationSeverity",
    "enforce_quantifier",
    "validate_batch_quantifiers",
    "validate_batch_typing",
    "validate_domain_range_alignment",
    "validate_entity_quantifiers",
]



"""Unit tests for CRM mapping layer."""

import pytest

from infoextract_cidoc.extraction.crm_mapper import map_to_crm_entities
from infoextract_cidoc.extraction.lite_schema import (
    LiteEntity,
    LiteExtractionResult,
    LiteRelationship,
)
from infoextract_cidoc.extraction.resolution import resolve_extraction
from infoextract_cidoc.models.base import CRMEntity, CRMRelation
from infoextract_cidoc.models.generated.e_classes import (
    E5_Event,
    E21_Person,
    E52_TimeSpan,
    E53_Place,
)


@pytest.fixture
def sample_extraction_result():
    lite_result = LiteExtractionResult(
        entities=[
            LiteEntity(ref_id="person_1", entity_type="Person", label="Einstein"),
            LiteEntity(ref_id="place_1", entity_type="Place", label="Ulm"),
        ],
        relationships=[
            LiteRelationship(
                source_ref="person_1",
                target_ref="place_1",
                property_code="P98",
                property_label="was born",
            ),
        ],
    )
    return resolve_extraction(lite_result)


@pytest.fixture
def event_extraction_result():
    """Fixture with an event linked to a timespan and place via P4/P7."""
    lite_result = LiteExtractionResult(
        entities=[
            LiteEntity(ref_id="event_1", entity_type="Event", label="Birth"),
            LiteEntity(ref_id="timespan_1", entity_type="TimeSpan", label="1879"),
            LiteEntity(ref_id="place_1", entity_type="Place", label="Ulm"),
        ],
        relationships=[
            LiteRelationship(
                source_ref="event_1",
                target_ref="timespan_1",
                property_code="P4",
                property_label="has time-span",
            ),
            LiteRelationship(
                source_ref="event_1",
                target_ref="place_1",
                property_code="P7",
                property_label="took place at",
            ),
        ],
    )
    return resolve_extraction(lite_result)


@pytest.mark.unit
class TestMapToCrmEntities:
    def test_returns_crm_entities(self, sample_extraction_result) -> None:
        entities, _relations = map_to_crm_entities(sample_extraction_result)
        assert len(entities) == 2
        assert all(isinstance(e, CRMEntity) for e in entities)

    def test_returns_crm_relations(self, sample_extraction_result) -> None:
        _entities, relations = map_to_crm_entities(sample_extraction_result)
        assert len(relations) == 1
        assert all(isinstance(r, CRMRelation) for r in relations)

    def test_entity_class_codes(self, sample_extraction_result) -> None:
        entities, _ = map_to_crm_entities(sample_extraction_result)
        codes = {e.class_code for e in entities}
        assert "E21" in codes
        assert "E53" in codes

    def test_typed_subclasses(self, sample_extraction_result) -> None:
        """Entities must be instances of the correct generated subclass."""
        entities, _ = map_to_crm_entities(sample_extraction_result)
        by_code = {e.class_code: e for e in entities}
        assert isinstance(by_code["E21"], E21_Person)
        assert isinstance(by_code["E53"], E53_Place)

    def test_relation_property_code(self, sample_extraction_result) -> None:
        _, relations = map_to_crm_entities(sample_extraction_result)
        assert relations[0].type == "P98"

    def test_entity_labels_preserved(self, sample_extraction_result) -> None:
        entities, _ = map_to_crm_entities(sample_extraction_result)
        labels = {e.label for e in entities}
        assert "Einstein" in labels
        assert "Ulm" in labels

    def test_shortcut_fields_populated(self, event_extraction_result) -> None:
        """P4 and P7 relationships must populate timespan and took_place_at."""
        entities, _ = map_to_crm_entities(event_extraction_result)
        by_code = {e.class_code: e for e in entities}
        event = by_code["E5"]
        timespan = by_code["E52"]
        place = by_code["E53"]
        assert isinstance(event, E5_Event)
        assert isinstance(timespan, E52_TimeSpan)
        assert event.timespan == timespan.id
        assert event.took_place_at == place.id

    def test_empty_result(self) -> None:
        from infoextract_cidoc.extraction.models import ExtractionResult

        empty = ExtractionResult(entities=[], relationships=[])
        entities, relations = map_to_crm_entities(empty)
        assert entities == []
        assert relations == []



"""
Base CRM entity model and core wrappers.
Provides the foundation for all CIDOC CRM E-class models.
"""

import hashlib
from typing import Any
from uuid import UUID, uuid4

from pydantic import BaseModel, ConfigDict, Field, field_validator

_UUID_NAMESPACE = UUID("6ba7b810-9dad-11d1-80b4-00c04fd430c8")  # DNS namespace


def _deterministic_uuid(value: str) -> UUID:
    """Return a deterministic UUID derived from *value* via MD5."""
    return UUID(
        hashlib.md5(
            f"{_UUID_NAMESPACE}{value}".encode(), usedforsecurity=False
        ).hexdigest()
    )


class CRMEntity(BaseModel):
    """
    Base class for all CIDOC CRM entities.

    Every CRM entity has:
    - id: unique identifier (UUID)
    - class_code: E-number (e.g., "E22")
    - label: human-readable name
    - notes: additional textual information
    - type: list of type assignments
    """

    model_config = ConfigDict(
        json_schema_extra={
            "description": "Base CIDOC CRM entity",
            "examples": [
                {
                    "id": "obj_001",
                    "class_code": "E22",
                    "label": "Ancient Vase",
                    "type": ["E55:Vessel", "E55:Ceramic"],
                }
            ],
        }
    )

    id: UUID = Field(
        default_factory=uuid4, description="Unique identifier for this entity"
    )
    class_code: str = Field(..., description="CIDOC CRM E-class code")
    label: str | None = Field(None, description="Human-readable label")
    notes: str | None = Field(None, description="Additional textual notes")
    source_text: str | None = Field(
        None,
        description="Source text snippet from which this entity was extracted",
    )
    type: list[str] = Field(default_factory=list, description="Type assignments")

    @field_validator("id", mode="before")
    @classmethod
    def convert_string_to_uuid(cls, v: Any) -> Any:
        """Convert string IDs to UUIDs for backward compatibility."""
        if isinstance(v, str):
            try:
                return UUID(v)
            except ValueError:
                return _deterministic_uuid(v)
        return v


class CRMRelation(BaseModel):
    """
    Represents a relationship between two CRM entities.

    Used internally for relationship expansion and Cypher emission.
    """

    model_config = ConfigDict(
        json_schema_extra={
            "description": "CRM relationship between entities",
            "examples": [
                {
                    "src": "obj_001",
                    "type": "P108",
                    "tgt": "prod_001",
                    "props": {"role": "E55:Painter"},
                }
            ],
        }
    )

    src: UUID = Field(..., description="Source entity ID")
    type: str = Field(..., description="P-property code (e.g., 'P108')")
    tgt: UUID = Field(..., description="Target entity ID")
    props: dict[str, Any] | None = Field(
        None, description="Additional relationship properties"
    )
    source_text: str | None = Field(
        None,
        description="Source text snippet from which this relationship was extracted",
    )

    @field_validator("src", "tgt", mode="before")
    @classmethod
    def convert_string_to_uuid(cls, v: Any) -> Any:
        """Convert string IDs to UUIDs for backward compatibility."""
        if isinstance(v, str):
            try:
                return UUID(v)
            except ValueError:
                return _deterministic_uuid(v)
        return v


class CRMValidationError(Exception):
    """Raised when CRM validation rules are violated."""


class CRMValidationWarning(Warning):
    """Issued when CRM validation rules are violated but severity is set to warn."""



"""
Unit tests for CRM validators.
"""

from uuid import uuid4

import pytest

from ...models.base import CRMValidationError
from ...models.generated.e_classes import (
    E12_Production,
    E22_HumanMadeObject,
)
from ...validators.quantifiers import (
    ValidationSeverity,
    enforce_quantifier,
    validate_entity_quantifiers,
)
from ...validators.typing_rules import (
    validate_batch_typing,
    validate_domain_range_alignment,
)


@pytest.mark.unit
class TestQuantifierValidation:
    """Test quantifier validation functionality."""

    def test_enforce_quantifier_valid(self):
        """Test quantifier enforcement with valid values."""
        entity = E22_HumanMadeObject(id=uuid4(), class_code="E22")

        # P108 has quantifier "0..1" - should allow 0 or 1 values
        enforce_quantifier(entity, "P108", [], ValidationSeverity.WARN)
        enforce_quantifier(entity, "P108", [uuid4()], ValidationSeverity.WARN)

    def test_enforce_quantifier_too_many(self):
        """Test quantifier enforcement with too many values."""
        entity = E22_HumanMadeObject(id=uuid4(), class_code="E22")

        # P108 has quantifier "0..1" - should not allow 2 values
        with pytest.raises(CRMValidationError):
            enforce_quantifier(
                entity, "P108", [uuid4(), uuid4()], ValidationSeverity.RAISE
            )

    def test_enforce_quantifier_too_few(self):
        """Test quantifier enforcement with too few values."""
        entity = E12_Production(id=uuid4(), class_code="E12")

        # P108i has quantifier "0..*" - should allow 0 values
        enforce_quantifier(entity, "P108i", [], ValidationSeverity.WARN)

    def test_validate_entity_quantifiers(self):
        """Test entity quantifier validation."""
        entity = E22_HumanMadeObject(
            id=uuid4(),
            class_code="E22",
            produced_by=uuid4(),  # This should be valid
        )

        messages = validate_entity_quantifiers(entity, ValidationSeverity.WARN)
        # Should not have validation errors for a valid entity
        assert len(messages) == 0


@pytest.mark.unit
class TestTypingValidation:
    """Test typing validation functionality."""

    def test_validate_domain_range_alignment_valid(self):
        """Test domain/range alignment with valid entities."""
        source = E22_HumanMadeObject(id=uuid4(), class_code="E22")
        target = E12_Production(id=uuid4(), class_code="E12")

        # P108: E22 -> E12 should be valid
        validate_domain_range_alignment(source, target, "P108", ValidationSeverity.WARN)

    def test_validate_domain_range_alignment_invalid(self):
        """Test domain/range alignment with invalid entities."""
        source = E22_HumanMadeObject(id=uuid4(), class_code="E22")
        target = E22_HumanMadeObject(id=uuid4(), class_code="E22")

        # P108: E22 -> E12 should be invalid with E22 target
        with pytest.raises(CRMValidationError):
            validate_domain_range_alignment(
                source, target, "P108", ValidationSeverity.RAISE
            )

    def test_validate_batch_typing(self):
        """Test batch typing validation."""
        entities = [
            E22_HumanMadeObject(id=uuid4(), class_code="E22"),
            E12_Production(id=uuid4(), class_code="E12"),
        ]

        results = validate_batch_typing(entities, ValidationSeverity.WARN)
        # Should not have validation errors for valid entities
        assert len(results) == 0


@pytest.mark.unit
class TestValidationSeverity:
    """Test validation severity handling."""

    def test_warn_severity(self):
        """Test that WARN severity issues warnings but doesn't raise."""
        entity = E22_HumanMadeObject(id="obj_001", class_code="E22")

        # This should issue a warning but not raise an exception
        enforce_quantifier(
            entity, "P108", ["prod_001", "prod_002"], ValidationSeverity.WARN
        )

    def test_raise_severity(self):
        """Test that RAISE severity raises exceptions."""
        entity = E22_HumanMadeObject(id="obj_001", class_code="E22")

        # This should raise an exception
        with pytest.raises(CRMValidationError):
            enforce_quantifier(
                entity, "P108", ["prod_001", "prod_002"], ValidationSeverity.RAISE
            )

    def test_ignore_severity(self):
        """Test that IGNORE severity doesn't validate."""
        entity = E22_HumanMadeObject(id="obj_001", class_code="E22")

        # This should not raise an exception or issue warnings
        enforce_quantifier(
            entity, "P108", ["prod_001", "prod_002"], ValidationSeverity.IGNORE
        )



"""
Styling utilities for network visualizations.

This module provides functions for customizing the appearance
of network plots and creating consistent styling.
"""

from typing import Any

import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import networkx as nx


def get_node_colors(
    graph: nx.Graph,
    color_scheme: str = "class_code",
    *,
    custom_colors: dict[str, str] | None = None,
) -> list[str]:
    """
    Get node colors based on specified color scheme.

    Args:
        graph: NetworkX graph
        color_scheme: Color scheme to use ("class_code", "degree", "centrality", "custom")
        custom_colors: Custom color mapping for "custom" scheme

    Returns:
        List of colors for each node
    """
    colors = []

    if color_scheme == "class_code":
        color_map = {
            "E21": "#FF6B6B",  # Person - Red
            "E5": "#4ECDC4",  # Event - Teal
            "E53": "#45B7D1",  # Place - Blue
            "E22": "#96CEB4",  # Object - Green
            "E52": "#FFEAA7",  # Time - Yellow
            "E74": "#DDA0DD",  # Group - Plum
            "E42": "#98D8C8",  # Identifier - Mint
            "E35": "#F7DC6F",  # Title - Gold
        }

        for node in graph.nodes():
            node_data = graph.nodes[node]
            class_code = node_data.get("class_code", "Unknown")
            colors.append(color_map.get(class_code, "#CCCCCC"))

    elif color_scheme == "degree":
        degrees = dict(graph.degree())
        max_degree = max(degrees.values()) if degrees else 1

        for node in graph.nodes():
            degree = degrees.get(node, 0)
            # Use a colormap to map degree to color
            normalized_degree = degree / max_degree
            color = plt.cm.viridis(normalized_degree)  # type: ignore[attr-defined]
            colors.append(color)

    elif color_scheme == "centrality":
        # This would require centrality scores to be passed
        # For now, use a default color
        colors = ["#FF6B6B"] * len(graph.nodes())

    elif color_scheme == "custom" and custom_colors:
        for node in graph.nodes():
            node_data = graph.nodes[node]
            class_code = node_data.get("class_code", "Unknown")
            colors.append(custom_colors.get(class_code, "#CCCCCC"))

    else:
        # Default color
        colors = ["#FF6B6B"] * len(graph.nodes())

    return colors


def get_edge_colors(
    graph: nx.Graph,
    color_scheme: str = "property_code",
    *,
    custom_colors: dict[str, str] | None = None,
) -> list[str]:
    """
    Get edge colors based on specified color scheme.

    Args:
        graph: NetworkX graph
        color_scheme: Color scheme to use ("property_code", "uniform", "custom")
        custom_colors: Custom color mapping for "custom" scheme

    Returns:
        List of colors for each edge
    """
    colors = []

    if color_scheme == "property_code":
        color_map = {
            "P11": "#FF6B6B",  # had participant - Red
            "P7": "#4ECDC4",  # took place at - Teal
            "P53": "#45B7D1",  # has current location - Blue
            "P108": "#96CEB4",  # was produced by - Green
            "P4": "#FFEAA7",  # has time-span - Yellow
            "P69": "#DDA0DD",  # has association with - Plum
        }

        for edge in graph.edges(data=True):
            edge_data = edge[2]
            property_code = edge_data.get("property_code", "Unknown")
            colors.append(color_map.get(property_code, "#CCCCCC"))

    elif color_scheme == "uniform":
        colors = ["#888888"] * graph.number_of_edges()

    elif color_scheme == "custom" and custom_colors:
        for edge in graph.edges(data=True):
            edge_data = edge[2]
            property_code = edge_data.get("property_code", "Unknown")
            colors.append(custom_colors.get(property_code, "#CCCCCC"))

    else:
        colors = ["#888888"] * graph.number_of_edges()

    return colors


def get_node_sizes(
    graph: nx.Graph,
    size_scheme: str = "degree",
    *,
    base_size: int = 300,
    min_size: int = 50,
    max_size: int = 1000,
    size_multiplier: float = 1.0,
) -> list[int]:
    """
    Get node sizes based on specified size scheme.

    Args:
        graph: NetworkX graph
        size_scheme: Size scheme to use ("degree", "uniform", "centrality")
        base_size: Base size for nodes
        min_size: Minimum node size
        max_size: Maximum node size
        size_multiplier: Multiplier for size calculation

    Returns:
        List of sizes for each node
    """
    sizes = []

    if size_scheme == "degree":
        degrees = dict(graph.degree())
        max_degree = max(degrees.values()) if degrees else 1

        for node in graph.nodes():
            degree = degrees.get(node, 0)
            normalized_degree = degree / max_degree
            size = base_size + (normalized_degree * base_size * size_multiplier)
            size = max(min_size, min(max_size, int(size)))
            sizes.append(size)

    elif size_scheme == "uniform":
        sizes = [base_size] * len(graph.nodes())

    elif size_scheme == "centrality":
        # This would require centrality scores to be passed
        # For now, use base size
        sizes = [base_size] * len(graph.nodes())

    else:
        sizes = [base_size] * len(graph.nodes())

    return sizes


def get_layout_positions(
    graph: nx.Graph,
    layout: str = "spring",
    *,
    seed: int | None = None,
    iterations: int = 50,
) -> dict[str, Any]:
    """
    Get node positions using specified layout algorithm.

    Args:
        graph: NetworkX graph
        layout: Layout algorithm ("spring", "circular", "hierarchical", "random", "kamada_kawai")
        seed: Random seed for reproducible layouts
        iterations: Number of iterations for spring layout

    Returns:
        Dictionary mapping node IDs to (x, y) positions
    """
    if layout == "spring":
        return nx.spring_layout(graph, iterations=iterations, seed=seed)  # type: ignore[no-any-return]
    if layout == "circular":
        return nx.circular_layout(graph)  # type: ignore[no-any-return]
    if layout == "hierarchical":
        try:
            return nx.nx_agraph.graphviz_layout(graph, prog="dot")  # type: ignore[no-any-return]
        except ImportError:
            # Fallback to spring layout if graphviz is not available
            return nx.spring_layout(graph, iterations=iterations, seed=seed)  # type: ignore[no-any-return]
    elif layout == "random":
        return nx.random_layout(graph, seed=seed)  # type: ignore[no-any-return]
    elif layout == "kamada_kawai":
        return nx.kamada_kawai_layout(graph)  # type: ignore[no-any-return]
    elif layout == "spectral":
        return nx.spectral_layout(graph)  # type: ignore[no-any-return]
    else:
        return nx.spring_layout(graph, iterations=iterations, seed=seed)  # type: ignore[no-any-return]


def create_legend(
    ax: plt.Axes,
    graph: nx.Graph,
    legend_type: str = "nodes",
    *,
    title: str | None = None,
    loc: str = "upper right",
    fontsize: int = 10,
) -> None:
    """
    Create legend for network plot.

    Args:
        ax: Matplotlib axes object
        graph: NetworkX graph
        legend_type: Type of legend ("nodes", "edges", "both")
        title: Legend title
        loc: Legend location
        fontsize: Font size for legend

    Returns:
        None
    """
    legend_elements = []

    if legend_type in ["nodes", "both"]:
        # Create node legend
        node_color_map = {
            "E21": ("Person", "#FF6B6B"),
            "E5": ("Event", "#4ECDC4"),
            "E53": ("Place", "#45B7D1"),
            "E22": ("Object", "#96CEB4"),
            "E52": ("Time", "#FFEAA7"),
            "E74": ("Group", "#DDA0DD"),
            "E42": ("Identifier", "#98D8C8"),
            "E35": ("Title", "#F7DC6F"),
        }

        # Only include classes that exist in the graph
        existing_classes = set()
        for node in graph.nodes():
            node_data = graph.nodes[node]
            class_code = node_data.get("class_code", "Unknown")
            existing_classes.add(class_code)

        for class_code, (label, color) in node_color_map.items():
            if class_code in existing_classes:
                legend_elements.append(mpatches.Patch(color=color, label=label))

    if legend_type in ["edges", "both"]:
        # Create edge legend
        edge_color_map = {
            "P11": ("had participant", "#FF6B6B"),
            "P7": ("took place at", "#4ECDC4"),
            "P53": ("has current location", "#45B7D1"),
            "P108": ("was produced by", "#96CEB4"),
            "P4": ("has time-span", "#FFEAA7"),
            "P69": ("has association with", "#DDA0DD"),
        }

        # Only include properties that exist in the graph
        existing_properties = set()
        for edge in graph.edges(data=True):
            edge_data = edge[2]
            property_code = edge_data.get("property_code", "Unknown")
            existing_properties.add(property_code)

        for prop_code, (label, color) in edge_color_map.items():
            if prop_code in existing_properties:
                legend_elements.append(mpatches.Patch(color=color, label=label))

    if legend_elements:
        legend = ax.legend(
            handles=legend_elements,
            loc=loc,
            fontsize=fontsize,
            title=title,
            title_fontsize=fontsize + 2,
        )
        legend.get_frame().set_facecolor("white")
        legend.get_frame().set_alpha(0.8)


def get_color_palette(
    palette_name: str = "default",
    *,
    n_colors: int = 8,
) -> list[str]:
    """
    Get a color palette for network visualization.

    Args:
        palette_name: Name of the color palette
        n_colors: Number of colors to return

    Returns:
        List of hex color codes
    """
    palettes = {
        "default": [
            "#FF6B6B",
            "#4ECDC4",
            "#45B7D1",
            "#96CEB4",
            "#FFEAA7",
            "#DDA0DD",
            "#98D8C8",
            "#F7DC6F",
        ],
        "pastel": [
            "#FFB3BA",
            "#FFDFBA",
            "#FFFFBA",
            "#BAFFC9",
            "#BAE1FF",
            "#E6BAFF",
            "#FFBAE6",
            "#FFE6BA",
        ],
        "vibrant": [
            "#FF0000",
            "#00FF00",
            "#0000FF",
            "#FFFF00",
            "#FF00FF",
            "#00FFFF",
            "#FFA500",
            "#800080",
        ],
        "muted": [
            "#8B4513",
            "#2F4F4F",
            "#483D8B",
            "#556B2F",
            "#8B008B",
            "#B22222",
            "#228B22",
            "#4682B4",
        ],
        "grayscale": [
            "#000000",
            "#333333",
            "#666666",
            "#999999",
            "#CCCCCC",
            "#FFFFFF",
            "#808080",
            "#404040",
        ],
    }

    palette = palettes.get(palette_name, palettes["default"])

    # Repeat or truncate palette to get desired number of colors
    if len(palette) >= n_colors:
        return palette[:n_colors]
    return [palette[i % len(palette)] for i in range(n_colors)]


def apply_style(
    ax: plt.Axes,
    *,
    background_color: str = "white",
    grid: bool = False,
    grid_color: str = "lightgray",
    grid_alpha: float = 0.3,
    border_color: str = "black",
    border_width: float = 1.0,
) -> None:
    """
    Apply styling to matplotlib axes.

    Args:
        ax: Matplotlib axes object
        background_color: Background color
        grid: Whether to show grid
        grid_color: Grid color
        grid_alpha: Grid transparency
        border_color: Border color
        border_width: Border width

    Returns:
        None
    """
    ax.set_facecolor(background_color)

    if grid:
        ax.grid(True, color=grid_color, alpha=grid_alpha, linestyle="-", linewidth=0.5)

    # Set border
    for spine in ax.spines.values():
        spine.set_color(border_color)
        spine.set_linewidth(border_width)

    # Remove ticks
    ax.set_xticks([])
    ax.set_yticks([])



"""
Network plotting functions for CIDOC CRM graphs.

This module provides various plotting functions for visualizing
NetworkX graphs containing CRM entities and relationships.
"""

from typing import Any

import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import plotly.graph_objects as go


def plot_network_graph(
    graph: nx.Graph,
    *,
    title: str = "CRM Network",
    figsize: tuple[int, int] = (12, 8),
    layout: str = "spring",
    node_size: int = 300,
    node_color: str = "class_code",
    edge_width: float = 1.0,
    edge_color: str = "gray",
    show_labels: bool = True,
    label_font_size: int = 8,
    save_path: str | None = None,
    show_plot: bool = True,
) -> plt.Figure:
    """
    Create a static network plot using matplotlib.

    Args:
        graph: NetworkX graph to plot
        title: Plot title
        figsize: Figure size (width, height)
        layout: Layout algorithm ("spring", "circular", "hierarchical", "random")
        node_size: Base node size
        node_color: Node coloring scheme ("class_code", "degree", "centrality")
        edge_width: Edge width
        edge_color: Edge color
        show_labels: Whether to show node labels
        label_font_size: Font size for labels
        save_path: Path to save the plot
        show_plot: Whether to display the plot

    Returns:
        Matplotlib figure object
    """
    fig, ax = plt.subplots(figsize=figsize)

    # Get layout positions
    pos = _get_layout_positions(graph, layout)

    # Get node colors and sizes
    node_colors = _get_node_colors(graph, node_color)
    node_sizes = _get_node_sizes(graph, node_size)

    # Draw the graph
    nx.draw(
        graph,
        pos=pos,
        ax=ax,
        node_color=node_colors,
        node_size=node_sizes,
        edge_color=edge_color,
        width=edge_width,
        with_labels=show_labels,
        font_size=label_font_size,
        font_weight="bold",
        alpha=0.8,
    )

    # Add title
    ax.set_title(title, fontsize=16, fontweight="bold", pad=20)

    # Create legend
    _create_legend(ax, graph, node_color)

    # Adjust layout
    plt.tight_layout()

    # Save if requested
    if save_path:
        plt.savefig(save_path, dpi=300, bbox_inches="tight")

    # Show if requested
    if show_plot:
        plt.show()

    return fig


def plot_temporal_network(
    graph: nx.Graph,
    *,
    title: str = "Temporal CRM Network",
    time_attribute: str = "temporal_info",
    figsize: tuple[int, int] = (14, 10),
    show_timeline: bool = True,
    save_path: str | None = None,
) -> plt.Figure:
    """
    Create a temporal network plot showing time-based relationships.

    Args:
        graph: NetworkX graph with temporal attributes
        title: Plot title
        time_attribute: Name of temporal attribute
        figsize: Figure size
        show_timeline: Whether to show timeline axis
        save_path: Path to save the plot

    Returns:
        Matplotlib figure object
    """
    fig, ax = plt.subplots(figsize=figsize)

    # Separate nodes by temporal information
    temporal_nodes = []
    non_temporal_nodes = []

    for node_id in graph.nodes():
        node_data = graph.nodes[node_id]
        if node_data.get("has_temporal_info", False):
            temporal_nodes.append(node_id)
        else:
            non_temporal_nodes.append(node_id)

    # Create subgraph for temporal nodes
    pos: dict[str, Any] = {}
    if temporal_nodes:
        temporal_subgraph = graph.subgraph(temporal_nodes)

        # Use time-based layout
        pos = _get_temporal_layout(temporal_subgraph, time_attribute)

        # Draw temporal nodes
        nx.draw_networkx_nodes(
            temporal_subgraph,
            pos=pos,
            ax=ax,
            node_color="red",
            node_size=400,
            alpha=0.8,
            label="Temporal Entities",
        )

        # Draw temporal edges
        nx.draw_networkx_edges(
            temporal_subgraph, pos=pos, ax=ax, edge_color="red", width=1.5, alpha=0.6
        )

    # Draw non-temporal nodes
    if non_temporal_nodes:
        non_temporal_subgraph = graph.subgraph(non_temporal_nodes)
        pos_non_temporal = nx.spring_layout(non_temporal_subgraph)

        nx.draw_networkx_nodes(
            non_temporal_subgraph,
            pos=pos_non_temporal,
            ax=ax,
            node_color="lightblue",
            node_size=300,
            alpha=0.6,
            label="Non-Temporal Entities",
        )

    # Add labels
    nx.draw_networkx_labels(graph, pos, ax=ax, font_size=8)

    ax.set_title(title, fontsize=16, fontweight="bold")
    ax.legend()

    if show_timeline:
        _add_timeline_axis(ax, graph, time_attribute)

    plt.tight_layout()

    if save_path:
        plt.savefig(save_path, dpi=300, bbox_inches="tight")

    return fig


def plot_community_network(
    graph: nx.Graph,
    communities: list[list[str]],
    *,
    title: str = "Community Structure",
    figsize: tuple[int, int] = (12, 8),
    save_path: str | None = None,
) -> plt.Figure:
    """
    Plot network with community structure highlighted.

    Args:
        graph: NetworkX graph
        communities: List of communities (lists of node IDs)
        title: Plot title
        figsize: Figure size
        save_path: Path to save the plot

    Returns:
        Matplotlib figure object
    """
    fig, ax = plt.subplots(figsize=figsize)

    # Get layout
    pos = nx.spring_layout(graph)

    # Color nodes by community
    colors = plt.cm.Set3(np.linspace(0, 1, len(communities)))  # type: ignore[attr-defined]

    for i, community in enumerate(communities):
        community_subgraph = graph.subgraph(community)

        nx.draw_networkx_nodes(
            community_subgraph,
            pos=pos,
            ax=ax,
            node_color=[colors[i]],
            node_size=400,
            alpha=0.8,
            label=f"Community {i + 1}",
        )

    # Draw edges
    nx.draw_networkx_edges(
        graph, pos=pos, ax=ax, edge_color="gray", width=0.5, alpha=0.5
    )

    # Draw labels
    nx.draw_networkx_labels(graph, pos, ax=ax, font_size=8)

    ax.set_title(title, fontsize=16, fontweight="bold")
    ax.legend()

    plt.tight_layout()

    if save_path:
        plt.savefig(save_path, dpi=300, bbox_inches="tight")

    return fig


def plot_centrality_network(
    graph: nx.Graph,
    centrality_scores: dict[str, float],
    *,
    title: str = "Centrality Analysis",
    centrality_type: str = "degree",
    figsize: tuple[int, int] = (12, 8),
    save_path: str | None = None,
) -> plt.Figure:
    """
    Plot network with node sizes based on centrality scores.

    Args:
        graph: NetworkX graph
        centrality_scores: Dictionary of node centrality scores
        title: Plot title
        centrality_type: Type of centrality being visualized
        figsize: Figure size
        save_path: Path to save the plot

    Returns:
        Matplotlib figure object
    """
    fig, ax = plt.subplots(figsize=figsize)

    # Get layout
    pos = nx.spring_layout(graph)

    # Scale node sizes based on centrality
    max_centrality = max(centrality_scores.values()) if centrality_scores else 1
    node_sizes = [
        centrality_scores.get(node, 0) / max_centrality * 1000 + 100
        for node in graph.nodes()
    ]

    # Draw nodes
    nx.draw_networkx_nodes(
        graph, pos=pos, ax=ax, node_size=node_sizes, node_color="lightcoral", alpha=0.8
    )

    # Draw edges
    nx.draw_networkx_edges(
        graph, pos=pos, ax=ax, edge_color="gray", width=0.5, alpha=0.5
    )

    # Draw labels
    nx.draw_networkx_labels(graph, pos, ax=ax, font_size=8)

    ax.set_title(
        f"{title} - {centrality_type.title()} Centrality",
        fontsize=16,
        fontweight="bold",
    )

    plt.tight_layout()

    if save_path:
        plt.savefig(save_path, dpi=300, bbox_inches="tight")

    return fig


def create_interactive_plot(
    graph: nx.Graph,
    *,
    title: str = "Interactive CRM Network",
    layout: str = "spring",
    show_labels: bool = True,
    node_size_multiplier: float = 20,
    edge_width_multiplier: float = 2,
) -> go.Figure:
    """
    Create an interactive network plot using Plotly for notebook display.

    Note: This function is designed for use in Jupyter notebooks only.
    It creates interactive visualizations that work well in notebook environments.

    Args:
        graph: NetworkX graph to plot
        title: Plot title
        layout: Layout algorithm
        show_labels: Whether to show node labels
        node_size_multiplier: Multiplier for node sizes
        edge_width_multiplier: Multiplier for edge widths

    Returns:
        Plotly figure object for notebook display
    """
    # Get layout positions
    pos = _get_layout_positions(graph, layout)

    # Prepare edge traces
    edge_x = []
    edge_y = []

    for edge in graph.edges():
        x0, y0 = pos[edge[0]]
        x1, y1 = pos[edge[1]]
        edge_x.extend([x0, x1, None])
        edge_y.extend([y0, y1, None])

    edge_trace = go.Scatter(
        x=edge_x,
        y=edge_y,
        line={"width": edge_width_multiplier, "color": "#888"},
        hoverinfo="none",
        mode="lines",
    )

    # Prepare node traces
    node_x = []
    node_y = []
    node_text = []
    node_hovertext = []
    node_colors = []
    node_sizes = []

    for node in graph.nodes():
        x, y = pos[node]
        node_x.append(x)
        node_y.append(y)

        node_data = graph.nodes[node]
        label = node_data.get("label", node)
        class_code = node_data.get("class_code", "Unknown")

        node_text.append(label if show_labels else "")
        node_hovertext.append(f"ID: {node}<br>Label: {label}<br>Class: {class_code}")

        # Color by class code
        color_map = {
            "E21": "red",  # Person
            "E5": "blue",  # Event
            "E53": "green",  # Place
            "E22": "orange",  # Object
            "E52": "purple",  # Time
        }
        node_colors.append(color_map.get(class_code, "gray"))

        # Size by degree
        degree = graph.degree(node)
        node_sizes.append(max(10, degree * node_size_multiplier))

    node_trace = go.Scatter(
        x=node_x,
        y=node_y,
        mode="markers+text",
        hoverinfo="text",
        hovertext=node_hovertext,
        text=node_text,
        textposition="middle center",
        marker={
            "size": node_sizes,
            "color": node_colors,
            "line": {"width": 2, "color": "white"},
        },
    )

    # Create figure
    return go.Figure(
        data=[edge_trace, node_trace],
        layout=go.Layout(
            title={"text": title, "font": {"size": 16}},
            showlegend=False,
            hovermode="closest",
            margin={"b": 20, "l": 5, "r": 5, "t": 40},
            annotations=[
                {
                    "text": "Interactive CRM Network Visualization",
                    "showarrow": False,
                    "xref": "paper",
                    "yref": "paper",
                    "x": 0.005,
                    "y": -0.002,
                    "xanchor": "left",
                    "yanchor": "bottom",
                    "font": {"color": "#888", "size": 12},
                }
            ],
            xaxis={"showgrid": False, "zeroline": False, "showticklabels": False},
            yaxis={"showgrid": False, "zeroline": False, "showticklabels": False},
        ),
    )


# Helper functions


def _get_layout_positions(graph: nx.Graph, layout: str) -> dict[str, Any]:
    """Get node positions using specified layout algorithm."""
    if layout == "spring":
        return nx.spring_layout(graph)  # type: ignore[no-any-return]
    if layout == "circular":
        return nx.circular_layout(graph)  # type: ignore[no-any-return]
    if layout == "hierarchical":
        return nx.nx_agraph.graphviz_layout(graph, prog="dot")  # type: ignore[no-any-return]
    if layout == "random":
        return nx.random_layout(graph)  # type: ignore[no-any-return]
    return nx.spring_layout(graph)  # type: ignore[no-any-return]


def _get_node_colors(graph: nx.Graph, color_scheme: str) -> list[str]:
    """Get node colors based on specified scheme."""
    colors = []
    color_map = {
        "E21": "red",  # Person
        "E5": "blue",  # Event
        "E53": "green",  # Place
        "E22": "orange",  # Object
        "E52": "purple",  # Time
    }

    for node in graph.nodes():
        node_data = graph.nodes[node]
        class_code = node_data.get("class_code", "Unknown")
        colors.append(color_map.get(class_code, "gray"))

    return colors


def _get_node_sizes(graph: nx.Graph, base_size: int) -> list[int]:
    """Get node sizes based on degree."""
    sizes = []
    degrees = dict(graph.degree())
    max_degree = max(degrees.values()) if degrees else 1

    # Handle case where all nodes have degree 0
    if max_degree == 0:
        return [base_size] * graph.number_of_nodes()

    for node in graph.nodes():
        degree = degrees.get(node, 0)
        size = base_size + (degree / max_degree) * base_size
        sizes.append(int(size))

    return sizes


def _create_legend(ax: plt.Axes, graph: nx.Graph, color_scheme: str) -> None:
    """Create legend for the plot."""
    if color_scheme == "class_code":
        legend_elements = []
        color_map = {
            "E21": ("Person", "red"),
            "E5": ("Event", "blue"),
            "E53": ("Place", "green"),
            "E22": ("Object", "orange"),
            "E52": ("Time", "purple"),
        }

        for class_code, (label, color) in color_map.items():
            if any(
                graph.nodes[node].get("class_code") == class_code
                for node in graph.nodes()
            ):
                legend_elements.append(mpatches.Patch(color=color, label=label))

        if legend_elements:
            ax.legend(handles=legend_elements, loc="upper right")


def _get_temporal_layout(graph: nx.Graph, time_attribute: str) -> dict[str, Any]:
    """Get layout positions based on temporal information."""
    # This is a simplified implementation
    # In practice, you would parse temporal data and position nodes accordingly
    return nx.spring_layout(graph)  # type: ignore[no-any-return]


def _add_timeline_axis(ax: plt.Axes, graph: nx.Graph, time_attribute: str) -> None:
    """Add timeline axis to the plot."""
    # This would be implemented based on actual temporal data



"""Entity and relationship resolution pipeline.

Two-stage pipeline:
  Stage A: Register all LiteEntities, assign stable UUIDs, produce ExtractedEntity list
  Stage B: Resolve LiteRelationships against the registry, reject broken links
"""

from __future__ import annotations

import logging
import uuid
from typing import TYPE_CHECKING

from infoextract_cidoc.extraction.models import (
    EventExtraction,
    ExtractedEntity,
    ExtractedRelationship,
    ExtractionResult,
    ObjectExtraction,
    PersonExtraction,
    PlaceExtraction,
    TimeExtraction,
)

if TYPE_CHECKING:
    from infoextract_cidoc.extraction.lite_schema import (
        LiteEntity,
        LiteExtractionResult,
    )

logger = logging.getLogger(__name__)

# Stable namespace for UUID5 generation
_EXTRACTION_NAMESPACE = uuid.UUID("6ba7b810-9dad-11d1-80b4-00c04fd430c8")

# Map entity_type strings to ExtractedEntity subclasses and class_codes
_ENTITY_TYPE_MAP: dict[str, tuple[type[ExtractedEntity], str]] = {
    "Person": (PersonExtraction, "E21"),
    "Event": (EventExtraction, "E5"),
    "Place": (PlaceExtraction, "E53"),
    "Object": (ObjectExtraction, "E22"),
    "TimeSpan": (TimeExtraction, "E52"),
}

# Default values for required fields on ExtractedEntity subclasses
_SUBCLASS_DEFAULTS: dict[type[ExtractedEntity], dict[str, str]] = {
    EventExtraction: {"event_type": "unknown"},
    PlaceExtraction: {"place_type": "unknown"},
    ObjectExtraction: {"object_type": "unknown"},
    TimeExtraction: {"time_type": "unknown"},
}


def _stable_uuid(ref_id: str, label: str) -> uuid.UUID:
    """Generate a stable UUID5 from ref_id and label."""
    return uuid.uuid5(_EXTRACTION_NAMESPACE, f"{ref_id}:{label}")


def _make_extracted_entity(lite: LiteEntity) -> ExtractedEntity:
    """Convert a LiteEntity to an ExtractedEntity subclass."""
    entity_cls, class_code = _ENTITY_TYPE_MAP.get(
        lite.entity_type, (ExtractedEntity, "E55")
    )
    stable_id = _stable_uuid(lite.ref_id, lite.label)

    # Build kwargs, supplying defaults for required subclass fields
    kwargs: dict = {
        "id": stable_id,
        "class_code": class_code,
        "label": lite.label,
        "description": lite.description,
        "confidence": lite.confidence,
        "source_text": lite.source_snippet,
        "properties": lite.attributes,
    }

    # Merge defaults for required fields, allowing attributes to override
    defaults = _SUBCLASS_DEFAULTS.get(entity_cls, {})
    for field_name, default_value in defaults.items():
        kwargs[field_name] = lite.attributes.get(field_name, default_value)

    return entity_cls(**kwargs)


class EntityRegistry:
    """Maps ref_ids to stable ExtractedEntity instances."""

    def __init__(self) -> None:
        self._by_ref_id: dict[str, ExtractedEntity] = {}
        self._by_label: dict[str, ExtractedEntity] = {}

    def register(self, lite: LiteEntity) -> ExtractedEntity:
        """Register a LiteEntity, deduplicating by label if needed."""
        # Deduplication: if same label exists, return existing entity
        if lite.label in self._by_label:
            existing = self._by_label[lite.label]
            self._by_ref_id[lite.ref_id] = existing
            return existing

        entity = _make_extracted_entity(lite)
        self._by_ref_id[lite.ref_id] = entity
        self._by_label[lite.label] = entity
        return entity

    def resolve(self, ref_id: str) -> ExtractedEntity | None:
        """Resolve a ref_id to an ExtractedEntity, or None if not found."""
        return self._by_ref_id.get(ref_id)

    @property
    def entities(self) -> list[ExtractedEntity]:
        """Return all unique registered entities."""
        # Use label dedup: _by_label values are unique
        seen_ids: set[uuid.UUID] = set()
        result: list[ExtractedEntity] = []
        for entity in self._by_label.values():
            if entity.id not in seen_ids:
                seen_ids.add(entity.id)
                result.append(entity)
        return result


def resolve_extraction(lite_result: LiteExtractionResult) -> ExtractionResult:
    """Resolve a LiteExtractionResult into a full ExtractionResult.

    Stage A: Register all entities, assign stable UUIDs.
    Stage B: Resolve relationships, log and exclude broken links.

    Args:
        lite_result: The raw LangStruct extraction output.

    Returns:
        ExtractionResult with stable UUIDs and resolved relationships.
    """
    registry = EntityRegistry()

    # Stage A: Register all entities
    for lite_entity in lite_result.entities:
        registry.register(lite_entity)

    entities = registry.entities

    # Stage B: Resolve relationships
    relationships: list[ExtractedRelationship] = []
    for lite_rel in lite_result.relationships:
        source = registry.resolve(lite_rel.source_ref)
        target = registry.resolve(lite_rel.target_ref)

        if source is None:
            logger.warning(
                "Broken relationship: source_ref %r not found in registry. Skipping.",
                lite_rel.source_ref,
            )
            continue

        if target is None:
            logger.warning(
                "Broken relationship: target_ref %r not found in registry. Skipping.",
                lite_rel.target_ref,
            )
            continue

        relationships.append(
            ExtractedRelationship(
                source_id=source.id,
                target_id=target.id,
                property_code=lite_rel.property_code,
                property_label=lite_rel.property_label,
                confidence=lite_rel.confidence,
                source_text=lite_rel.source_snippet,
            )
        )

    return ExtractionResult(entities=entities, relationships=relationships)



"""
Golden test for museum object example.
Tests the complete workflow from JSON to Markdown to Cypher.
"""

import json
from pathlib import Path

import pytest

from ...io.to_cypher import generate_cypher_parameters, generate_cypher_script
from ...io.to_markdown import MarkdownStyle, render_table, to_markdown
from ...models.generated.e_classes import (
    E1_CRMEntity,
    E12_Production,
    E22_HumanMadeObject,
    E52_TimeSpan,
    E53_Place,
    E61_TimePrimitive,
)
from ...validators.quantifiers import ValidationSeverity, validate_batch_quantifiers
from ...validators.typing_rules import validate_batch_typing


@pytest.mark.golden
def test_museum_object_workflow():
    """Test the complete museum object workflow."""
    # Load the example data
    example_file = (
        Path(__file__).parent.parent.parent / "examples" / "museum_object.json"
    )
    with example_file.open() as f:
        data = json.load(f)

    # Create entities from JSON
    entities = []
    for entity_data in data["entities"]:
        if entity_data["class_code"] == "E22":
            entity = E22_HumanMadeObject(**entity_data)
        elif entity_data["class_code"] == "E12":
            entity = E12_Production(**entity_data)
        elif entity_data["class_code"] == "E53":
            entity = E53_Place(**entity_data)
        elif entity_data["class_code"] == "E52":
            entity = E52_TimeSpan(**entity_data)
        elif entity_data["class_code"] == "E61":
            entity = E61_TimePrimitive(**entity_data)
        else:
            entity = E1_CRMEntity(**entity_data)

        entities.append(entity)

    # Test Markdown rendering
    vase = entities[0]  # E22_HumanMadeObject
    markdown_card = to_markdown(vase, MarkdownStyle.CARD)
    markdown_detailed = to_markdown(vase, MarkdownStyle.DETAILED)
    markdown_table = render_table(entities)
    markdown_narrative = to_markdown(
        entities[1], MarkdownStyle.NARRATIVE
    )  # E12_Production

    # Verify Markdown output contains expected elements
    assert "### E22 · Human-Made Object · Ancient Greek Vase" in markdown_card
    assert "**Location** (`current_location`):" in markdown_card
    assert "**Produced By** (`produced_by`):" in markdown_card

    assert "## E22 · Human-Made Object — Ancient Greek Vase" in markdown_detailed
    assert "**Notes** (`notes`): A beautifully preserved amphora" in markdown_detailed

    assert "| id | class_code | label | type |" in markdown_table
    assert (
        "| E22 | Ancient Greek Vase |" in markdown_table
        or "Ancient Greek Vase" in markdown_table
    )

    assert "**Vase Production**" in markdown_narrative
    assert "is a production" in markdown_narrative
    assert "that occurred during" in markdown_narrative

    # Test Cypher generation
    cypher_script = generate_cypher_script(entities)
    cypher_params = generate_cypher_parameters(entities)

    # Verify Cypher output contains expected elements
    assert "// Create constraints" in cypher_script
    assert (
        "CREATE CONSTRAINT crm_id IF NOT EXISTS FOR (n:CRM) REQUIRE n.id IS UNIQUE;"
        in cypher_script
    )
    assert "// Create nodes" in cypher_script
    assert "UNWIND $nodes_0 AS n" in cypher_script
    assert "MERGE (x:CRM {id: n.id})" in cypher_script
    assert "SET x.class_code = n.class_code" in cypher_script

    # Verify relationship creation
    assert "// Create relationships" in cypher_script
    assert "UNWIND $rels_P53_HAS_CURRENT_LOCATION_0 AS r" in cypher_script
    assert "MERGE (s)-[:`P53_HAS_CURRENT_LOCATION`]->(t);" in cypher_script

    # Verify parameters
    assert "nodes_0" in cypher_params
    assert len(cypher_params["nodes_0"]) == 7  # All 7 entities

    # Check specific node data
    node_data = cypher_params["nodes_0"]
    vase_node = next(node for node in node_data if node["class_code"] == "E22")
    assert vase_node["class_code"] == "E22"
    assert vase_node["label"] == "Ancient Greek Vase"
    assert vase_node["type"] == ["E55:Vessel", "E55:Ceramic"]

    # Check relationship parameters
    assert "rels_P53_HAS_CURRENT_LOCATION_0" in cypher_params
    assert "rels_P108_WAS_PRODUCED_BY_0" in cypher_params
    assert "rels_P4_HAS_TIME_SPAN_0" in cypher_params
    assert "rels_P7_TOOK_PLACE_AT_0" in cypher_params
    assert "rels_P79_BEGIN_OF_THE_BEGIN_0" in cypher_params
    assert "rels_P80_END_OF_THE_END_0" in cypher_params

    # Verify relationship data
    location_rels = cypher_params["rels_P53_HAS_CURRENT_LOCATION_0"]
    assert len(location_rels) == 1
    # Check that source is E22 entity and target is E53 entity
    assert location_rels[0]["src"] == vase_node["id"]

    production_rels = cypher_params["rels_P108_WAS_PRODUCED_BY_0"]
    assert len(production_rels) == 1
    assert production_rels[0]["src"] == vase_node["id"]


@pytest.mark.golden
def test_museum_object_validation():
    """Test validation of the museum object example."""
    # Load and create entities
    example_file = (
        Path(__file__).parent.parent.parent / "examples" / "museum_object.json"
    )
    with example_file.open() as f:
        data = json.load(f)

    entities = []
    for entity_data in data["entities"]:
        if entity_data["class_code"] == "E22":
            entity = E22_HumanMadeObject(**entity_data)
        elif entity_data["class_code"] == "E12":
            entity = E12_Production(**entity_data)
        elif entity_data["class_code"] == "E53":
            entity = E53_Place(**entity_data)
        elif entity_data["class_code"] == "E52":
            entity = E52_TimeSpan(**entity_data)
        elif entity_data["class_code"] == "E61":
            entity = E61_TimePrimitive(**entity_data)
        else:
            entity = E1_CRMEntity(**entity_data)

        entities.append(entity)

    # Test quantifier validation
    quantifier_results = validate_batch_quantifiers(entities, ValidationSeverity.WARN)
    assert len(quantifier_results) == 0  # Should have no validation issues

    # Test typing validation
    typing_results = validate_batch_typing(entities, ValidationSeverity.WARN)
    assert len(typing_results) == 0  # Should have no validation issues


@pytest.mark.golden
def test_museum_object_roundtrip():
    """Test roundtrip conversion from JSON to entities and back."""
    # Load the example data
    example_file = (
        Path(__file__).parent.parent.parent / "examples" / "museum_object.json"
    )
    with example_file.open() as f:
        original_data = json.load(f)

    # Create entities from JSON
    entities = []
    for entity_data in original_data["entities"]:
        if entity_data["class_code"] == "E22":
            entity = E22_HumanMadeObject(**entity_data)
        elif entity_data["class_code"] == "E12":
            entity = E12_Production(**entity_data)
        elif entity_data["class_code"] == "E53":
            entity = E53_Place(**entity_data)
        elif entity_data["class_code"] == "E52":
            entity = E52_TimeSpan(**entity_data)
        elif entity_data["class_code"] == "E61":
            entity = E61_TimePrimitive(**entity_data)
        else:
            entity = E1_CRMEntity(**entity_data)

        entities.append(entity)

    # Convert back to JSON
    converted_data = {"entities": [entity.model_dump() for entity in entities]}

    # Verify that key data is preserved
    assert len(converted_data["entities"]) == len(original_data["entities"])

    # Check specific entity (E22 Human-Made Object)
    original_vase = next(
        e for e in original_data["entities"] if e["class_code"] == "E22"
    )
    converted_vase = next(
        e for e in converted_data["entities"] if e["class_code"] == "E22"
    )

    assert original_vase["class_code"] == converted_vase["class_code"]
    assert original_vase["label"] == converted_vase["label"]
    assert original_vase["type"] == converted_vase["type"]
    # Convert UUID objects to strings for comparison
    assert str(original_vase["current_location"]) == str(
        converted_vase["current_location"]
    )
    assert str(original_vase["produced_by"]) == str(converted_vase["produced_by"])



"""
Markdown renderers for CIDOC CRM entities.
Provides card, detailed, table, and narrative rendering styles.
"""

from enum import Enum
from typing import Any
from uuid import UUID

from infoextract_cidoc.models.base import CRMEntity


class MarkdownStyle(str, Enum):
    """Available Markdown rendering styles."""

    CARD = "card"
    DETAILED = "detailed"
    TABLE = "table"
    NARRATIVE = "narrative"


def to_markdown(
    entity: CRMEntity,
    style: MarkdownStyle = MarkdownStyle.CARD,
    *,
    aliases: dict[str, str] | None = None,
    show_codes: bool = True,
) -> str:
    """
    Render a CRM entity to Markdown.

    Args:
        entity: The CRM entity to render
        style: Rendering style (card, detailed, table, narrative)
        aliases: Optional alias mapping for friendly names
        show_codes: Whether to show E/P codes in output

    Returns:
        Markdown string representation
    """
    if style == MarkdownStyle.CARD:
        return _render_card(entity, aliases, show_codes)
    if style == MarkdownStyle.DETAILED:
        return _render_detailed(entity, aliases, show_codes)
    if style == MarkdownStyle.TABLE:
        return _render_table([entity], aliases, show_codes)
    if style == MarkdownStyle.NARRATIVE:
        return _render_narrative(entity, aliases, show_codes)
    msg = f"Unknown style: {style}"
    raise ValueError(msg)


def render_table(
    entities: list[CRMEntity],
    columns: list[str] | None = None,
    *,
    aliases: dict[str, str] | None = None,
    show_codes: bool = True,
) -> str:
    """
    Render multiple entities as a Markdown table.

    Args:
        entities: List of CRM entities to render
        columns: Optional list of columns to include
        aliases: Optional alias mapping for friendly names
        show_codes: Whether to show E/P codes in output

    Returns:
        Markdown table string
    """
    return _render_table(entities, aliases, show_codes, columns)


def _render_card(
    entity: CRMEntity, aliases: dict[str, str] | None, show_codes: bool
) -> str:
    """Render entity as a card-style Markdown."""
    # Get friendly class name
    class_name = _get_friendly_class_name(entity.class_code, aliases)

    # Build header
    header_parts = [entity.class_code, class_name]
    if entity.label:
        header_parts.append(entity.label)
    header_parts.append(f"({_format_uuid_for_display(entity.id)})")

    header = "### " + " · ".join(header_parts)

    # Build body with canonical fields
    body_lines = []
    canonical_fields = _get_canonical_fields(entity)

    for field in canonical_fields:
        value = _get_field_value(entity, field)
        if value:
            friendly_name = _get_friendly_property_name(field, aliases)
            formatted_value = _format_uuid_for_display(value)
            if show_codes:
                body_lines.append(
                    f"- **{friendly_name}** (`{field}`): {formatted_value}"
                )
            else:
                body_lines.append(f"- **{friendly_name}**: {formatted_value}")

    # Add notes if present
    if entity.notes:
        body_lines.append(f"- **Notes**: {entity.notes}")

    result = header + "\n\n" + "\n".join(body_lines)

    # Add source attribution as italic blockquote
    if entity.source_text:
        result += f"\n\n> *{entity.source_text}*"

    return result


def _render_detailed(
    entity: CRMEntity, aliases: dict[str, str] | None, show_codes: bool
) -> str:
    """Render entity in detailed format with all non-empty fields."""
    # Get friendly class name
    class_name = _get_friendly_class_name(entity.class_code, aliases)

    # Build header
    header = f"## {entity.class_code} · {class_name}"
    if entity.label:
        header += f" — {entity.label}"
    header += f" ({_format_uuid_for_display(entity.id)})"

    # Build detailed body
    body_lines = []

    # Add all non-empty fields
    for field_name, field_value in entity.model_dump().items():
        if field_value and field_name not in ["id", "class_code"]:
            friendly_name = _get_friendly_property_name(field_name, aliases)
            formatted_value = _format_uuid_for_display(field_value)
            if show_codes:
                body_lines.append(
                    f"- **{friendly_name}** (`{field_name}`): {formatted_value}"
                )
            else:
                body_lines.append(f"- **{friendly_name}**: {formatted_value}")

    return header + "\n\n" + "\n".join(body_lines)


def _render_table(
    entities: list[CRMEntity],
    _aliases: dict[str, str] | None,
    _show_codes: bool,
    columns: list[str] | None = None,
) -> str:
    """Render entities as a Markdown table."""
    if not entities:
        return "No entities to display."

    # Determine columns
    if columns is None:
        columns = ["id", "class_code", "label", "type"]

    # Build header
    header_row = "| " + " | ".join(columns) + " |"
    separator_row = "| " + " | ".join(["---"] * len(columns)) + " |"

    # Build data rows
    data_rows = []
    for entity in entities:
        row_values = []
        for col in columns:
            value = _get_field_value(entity, col)
            if value is None:
                value = ""
            elif isinstance(value, list):
                value = ", ".join(_format_uuid_for_display(v) for v in value)
            else:
                value = _format_uuid_for_display(value)
            row_values.append(value)
        data_rows.append("| " + " | ".join(row_values) + " |")

    return "\n".join([header_row, separator_row, *data_rows])


def _render_narrative(
    entity: CRMEntity, aliases: dict[str, str] | None, _show_codes: bool
) -> str:
    """Render entity as a narrative (especially useful for events)."""
    # Get friendly class name
    class_name = _get_friendly_class_name(entity.class_code, aliases)

    # Build narrative
    narrative_parts = []

    if entity.label:
        narrative_parts.append(f"**{entity.label}**")

    narrative_parts.append(f"is a {class_name.lower()}")

    if entity.class_code in ["E5", "E7", "E8", "E12"]:
        # Event-specific narrative
        if hasattr(entity, "timespan") and entity.timespan:
            narrative_parts.append(
                f"that occurred during {_format_uuid_for_display(entity.timespan)}"
            )

        if hasattr(entity, "took_place_at") and entity.took_place_at:
            narrative_parts.append(
                f"at {_format_uuid_for_display(entity.took_place_at)}"
            )

    elif entity.class_code == "E22":
        # Human-made object narrative
        if hasattr(entity, "produced_by") and entity.produced_by:
            narrative_parts.append(
                f"that was produced by {_format_uuid_for_display(entity.produced_by)}"
            )

        if hasattr(entity, "current_location") and entity.current_location:
            narrative_parts.append(
                f"currently located at {_format_uuid_for_display(entity.current_location)}"
            )

    narrative = " ".join(narrative_parts) + "."

    # Add additional details
    if entity.notes:
        narrative += f"\n\n{entity.notes}"

    if entity.source_text:
        narrative += f"\n\n> *{entity.source_text}*"

    return narrative


def _get_friendly_class_name(class_code: str, aliases: dict[str, str] | None) -> str:
    """Get friendly class name from aliases or use default."""
    if aliases and class_code in aliases:
        return aliases[class_code]

    # Default mapping for common classes
    default_aliases = {
        "E1": "CRM Entity",
        "E5": "Event",
        "E7": "Activity",
        "E8": "Acquisition",
        "E12": "Production",
        "E21": "Person",
        "E22": "Human-Made Object",
        "E53": "Place",
        "E52": "Time-Span",
        "E42": "Identifier",
        "E35": "Title",
    }

    return default_aliases.get(class_code, class_code)


def _get_friendly_property_name(
    property_name: str, aliases: dict[str, str] | None
) -> str:
    """Get friendly property name from aliases or use default."""
    if aliases and property_name in aliases:
        return aliases[property_name]

    # Default mapping for common properties
    default_aliases = {
        "label": "Label",
        "type": "Type",
        "notes": "Notes",
        "source_text": "Source",
        "timespan": "Time-Span",
        "took_place_at": "Location",
        "current_location": "Location",
        "produced_by": "Produced By",
        "begin_of_the_begin": "Start",
        "end_of_the_end": "End",
    }

    return default_aliases.get(property_name, property_name.replace("_", " ").title())


def _get_canonical_fields(entity: CRMEntity) -> list[str]:
    """Get canonical fields for an entity based on its class."""
    # This would ideally come from the class metadata
    # For now, use a simple mapping
    canonical_mapping = {
        "E5": ["label", "type", "timespan", "took_place_at"],
        "E7": ["label", "type", "timespan", "took_place_at"],
        "E8": ["label", "type", "timespan", "took_place_at"],
        "E12": ["label", "type", "timespan", "took_place_at"],
        "E21": ["label", "type", "current_location"],
        "E22": ["label", "type", "current_location", "produced_by"],
        "E53": ["label", "type"],
        "E52": ["label", "type", "begin_of_the_begin", "end_of_the_end"],
        "E42": ["label", "type"],
        "E35": ["label", "type"],
    }

    return canonical_mapping.get(entity.class_code, ["label", "type"])


def _get_field_value(entity: CRMEntity, field_name: str) -> Any:
    """Get field value from entity, handling both direct attributes and shortcut fields."""
    if hasattr(entity, field_name):
        return getattr(entity, field_name)
    return None


def _format_uuid_for_display(uuid_value: Any) -> str:
    """Format UUID for display in Markdown."""
    if isinstance(uuid_value, UUID):
        # Show first 8 characters for readability
        return str(uuid_value)[:8] + "..."
    if isinstance(uuid_value, str):
        try:
            # Try to parse as UUID
            uuid_obj = UUID(uuid_value)
            return str(uuid_obj)[:8] + "..."
        except ValueError:
            # Not a UUID, return as-is
            return uuid_value
    return str(uuid_value)
